///////////////////////////////////////////////////////////////////////////
// DBMesh.h : by R.Lie
//

#ifndef _DBMesh_h_
#define _DBMesh_h_

#include <iostream>
#include <string>
#include <fstream>

#include "Geometry.h"

/**
 * This class provides interface to input data from the triangulation data file
 * generated by the grid generation software \p{bamg}. See reference from the
 * software to know the data format.
 */
class DBMesh : public SimplestMesh<2,2>
{
private:
	std::vector<int> bm;
public:
    DBMesh();
    virtual ~DBMesh();
public:
	void readData(const std::string&);
	virtual void generateMesh(Mesh<2,2>&);
	DeclException1(ExcMeshData, 
		char *, 
		<< "Mesh data error: " << arg1);
private:
	void readNode(std::istream&);
	void readTriangleElement(std::istream&);
	void readQuadrilateralElement(std::istream&);
};

#endif //_DBMesh_h_

//
// end of file
///////////////////////////////////////////////////////////////////////////
