/**
 * @file   SparseMatrixTool.h
 * @author Robert Lie
 * @date   Sun Nov  4 09:06:11 2007
 * 
 * @brief  
 * 
 * 
 */

#include <base/exceptions.h>
#include <lac/sparse_matrix.h>

#include "Miscellaneous.h"

/**
 * һЩϡϡģвĹߺ
 * 
 */
namespace SparseMatrixTool {

  /** 
   * ϡģ尴ˮƽӳΪһµģ
   * Ҫͬ
   * 
   * @param sp0 ߵģ
   * @param sp1 ұߵģ
   * @param sp õģ
   */
  void hCatSparsityPattern(const SparsityPattern& sp0,
			   const SparsityPattern& sp1,
			   SparsityPattern& sp); 

  /**
   * ϡģ尴ֱӳΪһµģ
   * Ҫͬ
   *
   * @param sp0 ϱߵģ
   * @param sp1 ±ߵģ
   * @param sp õģ
   */
  void vCatSparsityPattern(const SparsityPattern& sp0,
			   const SparsityPattern& sp1,
			   SparsityPattern& sp);

  /**
   * ϡģ尴ķʽӳΪһµ
   * ģ壬ҲǶڸϡ \f$ A \f$  \f$ B
   * \f$ǵõ
   *
   * \f[
   *    \left(\begin{array}{cc}
   *      A & B \\
   *      B^T & 0
   *    \end{array}\right)
   * \f]
   *
   * ģ塣Ҫ \f$ A \f$  \f$ B \f$ ͬ
   * Լ \f$ A \f$ һ
   * 
   * @param sp0  \f$ A \f$ ģ
   * @param sp1  \f$ B \f$ ģ
   * @param sp õģ
   */
  void gammaCatSparsityPattern(const SparsityPattern& sp0,
			       const SparsityPattern& sp1,
			       SparsityPattern& sp);

  /**
   * ϡģ尴նԽǷʽӳΪһµģ塣
   * 
   * @param sp0 ϽǾģ
   * @param sp1 ½Ǿģ
   * @param sp õģ
   */
  void dCatSparsityPattern(const SparsityPattern& sp0,
			   const SparsityPattern& sp1,
			   SparsityPattern& sp);

  /** 
   * ĸϡģȫΪһµģ
   *
   * \f[
   *    \left(\begin{array}{cc}
   *      A_{00} & A_{01} \\
   *      A_{01} & A_{11}
   *    \end{array}\right)
   * \f]
   * 
   * @param sp00 ϽǾģ
   * @param sp01 ϽǾģ
   * @param sp10 ½Ǿģ
   * @param sp11 ½Ǿģ
   * @param sp õģ
   */  
  void fullCatSparsityPattern(const SparsityPattern& sp00,
			      const SparsityPattern& sp01,
			      const SparsityPattern& sp10,
			      const SparsityPattern& sp11,
			      SparsityPattern& sp);

  /**
   * ϡˮƽʽӳΪһµϡ
   * 
   * @param m0 ߵľ
   * @param m1 ұߵľ
   * @param sp õ¾ģ
   * @param m õ¾
   * @param is_pattern_ok Ϊ棬ģй죻
   *                      ٶģѾȷĹ죻ȱ
   *                      ʡֵΪ档
   */
  template <typename number>
    void hCatSparseMatrix(const SparseMatrix<number>& m0,
			  const SparseMatrix<number>& m1,
			  SparsityPattern& sp,
			  SparseMatrix<number>& m,
			  bool is_pattern_ok = true);

  /**
   * ϡֱʽӳΪһµϡ
   * 
   * @param m0 ϱߵľ
   * @param m1 ±ߵľ
   * @param sp õ¾ģ
   * @param m õ¾
   * @param is_pattern_ok Ϊ棬ģй죻
   *                      ٶģѾȷĹ죻ȱ
   *                      ʡֵΪ档
   */
  template <typename number>
    void vCatSparseMatrix(const SparseMatrix<number>& m0,
			  const SparseMatrix<number>& m1,
			  SparsityPattern& sp,
			  SparseMatrix<number>& m,
			  bool is_pattern_ok = true);

  /**
   * ϡ󰴰ķʽӳΪһµϡ
   * ҲǶڸϡ \f$ A \f$  \f$ B
   * \f$ǵõ
   *
   * \f[
   *    \left(\begin{array}{cc}
   *      A & B \\
   *      B^T & 0
   *    \end{array}\right)
   * \f]
   *
   * @param m0  \f$ A \f$
   * @param m1  \f$ B \f$
   * @param sp õ¾ģ
   * @param m õ¾
   * @param is_pattern_ok Ϊ棬ģй죻
   *                      ٶģѾȷĹ죻ȱ
   *                      ʡֵΪ档
   */
  template <typename number>
    void gammaCatSparseMatrix(const SparseMatrix<number>& m0,
			      const SparseMatrix<number>& m1,
			      SparsityPattern& sp,
			      SparseMatrix<number>& m,
			      bool is_pattern_ok = true);

  /**
   * ϡնԽǷʽӳΪһµϡ
   * 
   * @param m0 ϽǾ
   * @param m1 ½Ǿ
   * @param sp õ¾ģ
   * @param m õ¾
   * @param is_pattern_ok Ϊ棬ģй죻
   *                      ٶģѾȷĹ죻ȱ
   *                      ʡֵΪ档
   */
  template <typename number>
    void dCatSparseMatrix(const SparseMatrix<number>& m0,
			  const SparseMatrix<number>& m1,
			  SparsityPattern& sp,
			  SparseMatrix<number>& m,
			  bool is_pattern_ok = true);

  /** 
   * ĸϡȫΪһµϡ
   *
   * \f[
   *    \left(\begin{array}{cc}
   *      A_{00} & A_{01} \\
   *      A_{10} & A_{11}
   *    \end{array}\right)
   * \f]
   * 
   * @param m00 ϽǾ
   * @param m01 ϽǾ
   * @param m10 ½Ǿ
   * @param m11 ½Ǿ
   * @param sp õ¾ģ
   * @param m õ¾
   * @param is_pattern_ok Ϊ棬ģй죻
   *                      ٶģѾȷĹ죻ȱ
   *                      ʡֵΪ档
   */  
  template <typename number>
    void fullCatSparseMatrix(const SparseMatrix<number>& m00,
			     const SparseMatrix<number>& m01,
			     const SparseMatrix<number>& m10,
			     const SparseMatrix<number>& m11,
			     SparsityPattern& sp,
			     SparseMatrix<number>& m,
			     bool is_pattern_ok = true);

  DeclException0(ExcNotCompressed);
  DeclException2(ExcDimensionDontMatch, int, int,
		 << "The dimensions " << arg1 << " and " << arg2
		 << " do not match properly.");
};

/**
 * end of file
 * 
 */
